

import pytest
from fpvgcc.fpv import process_map_file


EXAMPLE_FILES = {
    'tests/maps/example.arm-none-eabi.basic.0.map': None,
    'tests/maps/example.msp430-elf.basic.0.map': None,
    'tests/maps/example.msp430-elf.0.map': None,
}


@pytest.fixture(scope="module",
                params=EXAMPLE_FILES.keys())
def example_map(request):
    filename = request.param
    vectors = EXAMPLE_FILES[filename]
    return process_map_file(filename, profile='auto'), vectors




# Copyright (C) 2015 Quazar Technologies Pvt. Ltd.
#               2015 Chintalagiri Shashank
#
# This file is part of fpv-gcc.
#
# fpv-gcc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fpv-gcc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fpv-gcc.  If not, see <http://www.gnu.org/licenses/>.


import logging
from .ntree import NTreeNode, NTree


class SizeNTreeNode(NTreeNode):
    def __init__(self, parent=None, node_t=None):
        super(SizeNTreeNode, self).__init__(parent, node_t)
        self._leafsize = None

    @property
    def size(self):
        rval = 0
        if self.is_leaf:
            rval = self.leafsize
        else:
            for child in self.children:
                try:
                    rval += child.size
                except TypeError:
                    logging.warning("Size information not available for : "
                                    + child.gident)
                    return "Err"
        return rval

    @property
    def leafsize(self):
        raise NotImplementedError

    @leafsize.setter
    def leafsize(self, value):
        raise NotImplementedError


class SizeNTree(NTree):
    node_t = SizeNTreeNode

    def __init__(self):
        super(SizeNTree, self).__init__()

    @property
    def size(self):
        return self.root.size



# Copyright (c)
#   (c) 2015-16 Chintalagiri Shashank, Quazar Technologies Pvt. Ltd.
#
# This file is part of fpv-gcc.
#
# fpv-gcc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fpv-gcc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fpv-gcc.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import with_statement
from __future__ import print_function
from six import iteritems

import re
import logging

from .gccMemoryMap import GCCMemoryMap, MemoryRegion
from .profiles import get_profile
from .profiles.guess import guess_profile


class GCCMemoryMapParserSM(object):
    def __init__(self, ctx):
        self.ctx = ctx
        self.state = 'START'

        self.IDEP_STATE = 'START'
        self.idep_archive = None

        self.COMSYM_STATE = 'NORMAL'
        self.comsym_name = None

        self.LINKERMAP_STATE = 'NORMAL'
        self.linkermap_section = None
        self.linkermap_symbol = None
        self.linkermap_lastsymbol = None

        self.idep_archives = []
        self.idep_symbols = []
        self.common_symbols = []
        self.memory_regions = []

        self.loaded_files = []
        self.linker_defined_addresses = []
        self.memory_map = GCCMemoryMap(self.ctx)

    def __repr__(self):
        return \
            "<GCCMemoryMapParserSM \n" + \
            "  {0:>30} = {1}\n".format('STATE', self.state) + \
            "  {0:>30} = {1}\n".format('IDEP_STATE', self.IDEP_STATE) + \
            "  {0:>30} = {1}\n".format('idep_archive', self.idep_archive) + \
            "  {0:>30} = {1}\n".format('COMSYM_STATE', self.COMSYM_STATE) + \
            "    {0:>28} = {1}\n".format('comsym_name', self.comsym_name) + \
            "  {0:>30} = {1}\n" \
            "".format('LINKERMAP_STATE', self.LINKERMAP_STATE) + \
            "    {0:>28} = {1}\n" \
            "".format('linkermap_section', self.linkermap_section) + \
            "    {0:>28} = {1}\n" \
            "".format('linkermap_symbol', self.linkermap_symbol) + \
            "    {0:>28} = {1}\n" \
            "".format('linkermap_lastsymbol', self.linkermap_lastsymbol)


# Regular Expressions
re_headings = {
    'IN_DEPENDENCIES': re.compile(r"Archive member [\s\w]* file \(symbol\)"),
    'IN_COMMON_SYMBOLS': re.compile(r"Allocating common symbols"),
    'IN_DISCARDED_INPUT_SECTIONS': re.compile(r"Discarded input sections"),
    'IN_MEMORY_CONFIGURATION': re.compile(r"Memory Configuration"),
    'IN_LINKER_SCRIPT_AND_MEMMAP': re.compile(r"Linker script and memory map")}

re_b1_archive = re.compile(
    r'^(?P<folder>.*/)?(?P<archive>:|(.+?)(?:(\.[^.]*)))\((?P<symbol>.*)\)$'
)
re_b1_file = re.compile(
    r'^\s+(?P<folder>.*/)?(?P<file>:|(.+?)(?:(\.[^.]*)))\((?P<symbol>.*)\)$'
)
re_comsym_normal = re.compile(
    r'^(?P<symbol>\S*)\s+(?P<size>0[xX][0-9a-fA-F]+)\s+(?P<filefolder>.*/)?(?P<archivefile>:|.+?(?:\.[^.]*))\((?P<objfile>.*)\)$'  # noqa
)
re_comsym_nameonly = re.compile(
    r'^(?P<symbol>\S*)$'
)
re_comsym_detailonly = re.compile(
    r'^\s+(?P<size>0[xX][0-9a-fA-F]+)\s+(?P<filefolder>.*/)?(?P<archivefile>:|.+?(?:\.[^.]*))\((?P<objfile>.*)\)$'  # noqa
)

re_sectionelem = re.compile(
    r'^(?P<treepath>\.\S*)\s+(?P<address>0[xX][0-9a-fA-F]+)\s+(?P<size>0[xX][0-9a-fA-F]+)\s+(?P<filefolder>.*/)?(?P<archivefile>:|.+?(?:\.[^.]*))\((?P<objfile>.*)\)$'  # noqa
)
re_sectionelem_nameonly = re.compile(
    r'^(?P<symbol>\.\S*)$'
)
re_sectionelem_detailonly = re.compile(
    r'^\s+(?P<address>0[xX][0-9a-fA-F]+)\s+(?P<size>0[xX][0-9a-fA-F]+)\s+(?P<filefolder>.*/)?(?P<archivefile>:|.+?(?:\.[^.]*))\((?P<objfile>.*)\)$'  # noqa
)

re_memregion = re.compile(
    r'^(?P<region>\S*)\s+(?P<origin>0[xX][0-9a-fA-F]+)\s+(?P<size>0[xX][0-9a-fA-F]+)\s*(?P<attribs>.*)$'  # noqa
)

re_linkermap = {
    'LOAD': re.compile(
        r'^LOAD\s+(?P<filefolder>.*/)?(?P<file>:|.+?(?:\.[^.]*))$'),
    'DEFN_ADDR': re.compile(
        r'^\s+(?P<origin>0[xX][0-9a-fA-F]+)\s+(?P<name>.*)\s=\s+(?P<defn>0[xX][0-9a-fA-F]+)$'),  # noqa
    'SECTION_HEADINGS': re.compile(
        r'^(?P<name>[._]\S*)(?:\s+(?P<address>0[xX][0-9a-fA-F]+))?(?:\s+(?P<size>0[xX][0-9a-fA-F]+))?(?:\s+load address\s+(?P<loadaddress>0[xX][0-9a-fA-F]+))?$'),  # noqa
    'SYMBOL': re.compile(
        r'^\s(?P<name>\S+)(?:\s+(?P<address>0[xX][0-9a-fA-F]+))(?:\s+(?P<size>0[xX][0-9a-fA-F]+))\s+(?P<filefolder>.*/)?(?P<file>:|.+?(?:\.[^.\)]*))?(?:\((?P<file2>\S*)\))?$'),  # noqa
    'FILL': re.compile(
        r'^\s(?:\*fill\*)(?:\s+(?P<address>0[xX][0-9a-fA-F]+))(?:\s+(?P<size>0[xX][0-9a-fA-F]+))$'),  # noqa
    'SYMBOLONLY':
        re.compile(r'^\s(?P<name>[._]\S+)$'),
    'SYMBOLDETAIL': re.compile(
        r'^\s+(?:\s+(?P<address>0[xX][0-9a-fA-F]+))(?:\s+(?P<size>0[xX][0-9a-fA-F]+))\s+(?P<filefolder>.*/)?(?P<file>:|.+?(?:\.[^.\)]*))?(?:\((?P<file2>\S*)\))?$'),  # noqa
    'SECTIONDETAIL': re.compile(
        r'^\s+(?:\s+(?P<address>0[xX][0-9a-fA-F]+))(?:\s+(?P<size>0[xX][0-9a-fA-F]+))$'),  # noqa
    'SECTIONHEADINGONLY':
        re.compile(r'^(?P<name>[._]\S+)$'),
    'LINKALIASES': re.compile(
        r'^\s\*\(((?:[^\s\*\(\)]+(?:\*|)\s)*(?:[^\s\*\(\)]+(?:\*|)))\)$')  # noqa
}


def check_line_for_heading(l):
    for key, regex in iteritems(re_headings):
        if regex.match(l):
            logging.info("Entering File Region : " + key)
            return key
    return None


class IDLArchive(object):
    # Rough draft. Needs much work to make it usable
    def __init__(self, folder, archive, objfile):
        self.folder = folder
        self.archive = archive
        self.objfile = objfile
        self.becauseof = None

    def __repr__(self):
        r = "\n\nArchive in Dependencies : \n"
        r += self.archive + "\n"
        r += self.objfile + "\n"
        r += self.folder + "\n"
        r += repr(self.becauseof)
        return r


class IDLSymbol(object):
    # Rough draft. Needs much work to make it usable
    def __init__(self, folder, objfile, symbol):
        self.folder = folder
        self.objfile = objfile
        self.symbol = symbol

    def __repr__(self):
        r = "Because of :: \n"
        r += self.symbol + "\n"
        r += self.objfile + "\n"
        r += self.folder + "\n"
        return r


def process_dependencies_line(l, sm):
    # Rough draft. Needs much work to make it usable
    if sm.IDEP_STATE == 'START':
        res = re_b1_archive.findall(l)
        if len(res) and len(res[0]) == 5:
            archive = IDLArchive(res[0][1], res[0][1], res[0][4])
            sm.idep_archives.append(archive)
            sm.IDEP_STATE = 'ARCHIVE_DEFINED'
            sm.idep_archive = archive
    if sm.IDEP_STATE == 'ARCHIVE_DEFINED':
        res = re_b1_file.findall(l)
        if len(res) and len(res[0]) == 5:
            symbol = IDLSymbol(res[0][1], res[0][1], res[0][4])
            sm.idep_symbols.append(symbol)
            sm.IDEP_STATE = 'START'
            sm.idep_archive.becauseof = symbol


class CommonSymbol(object):
    def __init__(self, symbol, size, filefolder, archivefile, objfile):
        self.symbol = symbol
        self.size = int(size, 16)
        self.filefolder = filefolder
        self.archivefile = archivefile
        self.objfile = objfile

    def __repr__(self):
        r = '{0:.<30}{1:<8}{2:<20}{3:<40}{4}'.format(
            self.symbol, self.size or '', self.objfile or '',
            self.archivefile or '', self.filefolder or ''
        )
        return r


def process_common_symbols_line(l, sm):
    if sm.COMSYM_STATE == 'NORMAL':
        res = re_comsym_normal.findall(l)
        if len(res) and len(res[0]) == 5:
            sym = CommonSymbol(res[0][0], res[0][1], res[0][2],
                               res[0][3], res[0][4])
            sm.common_symbols.append(sym)
        else:
            res = re_comsym_nameonly.findall(l)
            if len(res) == 1:
                sm.comsym_name = res[0]
                sm.COMSYM_STATE = 'GOT_NAME'
    elif sm.COMSYM_STATE == 'GOT_NAME':
        res = re_comsym_detailonly.findall(l)
        if len(res) and len(res[0]) == 4:
            sym = CommonSymbol(sm.comsym_name, res[0][0], res[0][1],
                               res[0][2], res[0][3])
            sm.common_symbols.append(sym)
            sm.COMSYM_STATE = 'NORMAL'


def process_discarded_input_section_line(l, sm):
    pass


def process_memory_configuration_line(l, sm):
    res = re_memregion.findall(l)
    if len(res) and len(res[0]) == 4:
        region = MemoryRegion(res[0][0], res[0][1], res[0][2], res[0][3])
        sm.memory_map.memory_regions.append(region)


def process_linkermap_load_line(l, sm):
    res = re_linkermap['LOAD'].findall(l)
    if len(res) and len(res[0]) == 2:
        sm.loaded_files.append((res[0][0] + res[0][1]).strip())


class LinkerDefnAddr(object):
    def __init__(self, symbol, address, defn_addr):
        self.symbol = symbol
        self.address = int(address, 16)
        self.defn_addr = int(defn_addr, 16)

    def __repr__(self):
        r = self.symbol
        r += " :: " + hex(self.address)
        return r


def process_linkermap_defn_addr_line(l, sm):
    res = re_linkermap['DEFN_ADDR'].findall(l)
    if len(res) and len(res[0]) == 3:
        sm.linker_defined_addresses.append(
            LinkerDefnAddr(res[0][1], res[0][0], res[0][2])
        )


def linkermap_name_process(name, sm, checksection=True):
    name = name.strip()
    if name.startswith('_'):
        name = '.' + name
    if name.startswith('COMMON'):
        name = '.' + name
    if name.startswith('*fill*'):
        return '*fill*'
    if not name.startswith('.'):
        logging.error('Skipping : {0}'.format(name.rstrip()))
        return None
    name = sm.memory_map.aliases.encode(name)
    if checksection is False:
        return name
    if not name.startswith(sm.linkermap_section.gident):
        if name != sm.linkermap_section.gident:
            logging.warning("Possibly mismatched section : {0} ; {1}"
                            "".format(name, sm.linkermap_section.gident))
            name = sm.linkermap_section.gident + name
    return name


def linkermap_get_newnode(name, sm, allow_disambig=True,
                          objfile=None, at_fill=True):
    # print("getting new node :", name)
    newnode = sm.memory_map.get_node(name, create=True)
    if at_fill is True:
        if newnode.is_leaf_property_set or newnode._address is not None:
            # The node isn't a new one. The present data within it needs to
            # be handled first.

            # Push the current node into it's own leaf node. This is enough
            # for most cases.
            try:
                newnode.push_to_leaf()
            except TypeError:
                print("Error getting new node : {0}".format(name))
                raise
            except RuntimeError:
                print("Runtime Error getting new node : {0}".format(name))
                exit(0)

            # Now generate the new node as a child of the original target.
            # In some cases this needs the node disambiguation element
            # included.
            newnodename = objfile.replace('.', '_')
            newnodepath = "{0}.{1}".format(name, newnodename)

            # Check if disambiguation is already initialized or is recommended
            disambig = sm.memory_map.get_node_disambig(newnodepath,
                                                       prospective=True)

            # Accordingly modify name as needed
            if disambig is not None:
                newnodepath = "{0}:{1}".format(newnodepath, disambig + 1)

            # Get the new child node for what was originally requested
            newnode = linkermap_get_newnode(
                newnodepath, sm, allow_disambig=False,
                objfile=objfile, at_fill=True
            )
    return newnode


def process_linkermap_section_headings_line(l, sm):
    match = re_linkermap['SECTION_HEADINGS'].match(l)
    name = match.group('name').strip()
    name = linkermap_name_process(name, sm, False)
    if name is None:
        return
    if name == '*fill*':
        print("IN SH")
    newnode = linkermap_get_newnode(name, sm, True)
    if match.group('address') is not None:
        newnode.address = match.group('address').strip()
    if match.group('size') is not None:
        newnode.defsize = match.group('size').strip()
        if len(newnode.children) > 0:
            newnode = newnode.push_to_leaf()
    if match.group('address') is not None:
        sm.linkermap_section = newnode
        sm.LINKERMAP_STATE = 'IN_SECTION'
    else:
        sm.linkermap_section = newnode
        sm.LINKERMAP_STATE = 'GOT_SECTION_NAME'


def process_linkermap_section_heading_detail_line(l, sm):
    match = re_linkermap['SECTIONDETAIL'].match(l)
    newnode = sm.linkermap_section
    if match:
        if match.group('address') is not None:
            newnode.address = match.group('address').strip()
        if match.group('size') is not None:
            newnode.defsize = match.group('size').strip()
            if len(newnode.children) > 0:
                newnode.push_to_leaf()
    sm.LINKERMAP_STATE = 'IN_SECTION'


def process_linkermap_symbol_line(l, sm):
    if sm.linkermap_symbol is not None:
        logging.warning("Probably Missed Symbol Detail : " +
                        sm.linkermap_symbol)
        sm.linkermap_symbol = None
    match = re_linkermap['SYMBOL'].match(l)
    name = match.group('name').strip()
    name = linkermap_name_process(name, sm)
    if name is None:
        return
    if name == '*fill*':
        sm.linkermap_lastsymbol.fillsize = match.group('size').strip()
        return
    arfile = None
    objfile = None
    arfolder = None
    if match.group('file2') is not None:
        arfile = match.group('file').strip()
        objfile = match.group('file2').strip()
        if match.group('filefolder') is not None:
            arfolder = match.group('filefolder').strip()
    elif match.group('file') is not None:
        objfile = match.group('file').strip()
        if match.group('filefolder') is not None:
            arfolder = match.group('filefolder').strip()
    newnode = linkermap_get_newnode(name, sm, allow_disambig=True,
                                    objfile=objfile)
    if arfile is not None:
        newnode.arfile = arfile
    if objfile is not None:
        newnode.objfile = objfile
    if arfolder is not None:
        newnode.arfolder = arfolder
    if match.group('address') is not None:
        newnode.address = match.group('address').strip()
    if match.group('size') is not None:
        newnode.osize = match.group('size').strip()
        if len(newnode.children) > 0:
            newnode = newnode.push_to_leaf()
    sm.linkermap_lastsymbol = newnode


def process_linkermap_fill_line(l, sm):
    if sm.linkermap_symbol is not None:
        logging.warning("Probably Missed Symbol Detail : "
                        + sm.linkermap_symbol)
        sm.linkermap_symbol = None

    if sm.linkermap_lastsymbol is None or sm.linkermap_symbol is not None:
        logging.warning("Fill Container Unknown : " + l)
        return

    match = re_linkermap['FILL'].match(l)
    if match.group('size') is not None:
        sm.linkermap_lastsymbol.fillsize = int(match.group('size').strip(), 16)


def process_linkermap_symbolonly_line(l, sm):
    if sm.linkermap_symbol is not None:
        logging.warning("Probably Missed Symbol Detail : "
                        + sm.linkermap_symbol)
        sm.linkermap_symbol = None
    match = re_linkermap['SYMBOLONLY'].match(l)
    name = match.group('name').strip()
    name = linkermap_name_process(name, sm)
    if name is None:
        return
    if name == '*fill*':
        print("IN SO")
    sm.linkermap_symbol = name


def process_linkermap_section_detail_line(l, sm):
    match = re_linkermap['SYMBOLDETAIL'].match(l)
    name = sm.linkermap_symbol
    if name is None:
        return
    arfile = None
    objfile = None
    arfolder = None
    if match.group('file2') is not None:
        arfile = match.group('file').strip()
        objfile = match.group('file2').strip()
        if match.group('filefolder') is not None:
            arfolder = match.group('filefolder').strip()
    elif match.group('file') is not None:
        objfile = match.group('file').strip()
        if match.group('filefolder') is not None:
            arfolder = match.group('filefolder').strip()
    newnode = linkermap_get_newnode(name, sm,
                                    allow_disambig=True, objfile=objfile)
    if arfile is not None:
        newnode.arfile = arfile
    if objfile is not None:
        newnode.objfile = objfile
    if arfolder is not None:
        newnode.arfolder = arfolder
    if match.group('address') is not None:
        newnode.address = match.group('address').strip()
    if match.group('size') is not None:
        newnode.osize = match.group('size').strip()
        if len(newnode.children) > 0:
            newnode = newnode.push_to_leaf()
    sm.linkermap_lastsymbol = newnode
    sm.linkermap_symbol = None


def process_linkaliases_line(l, sm):
    match = re_linkermap['LINKALIASES'].match(l)
    alias_list = match.group(1).split(' ')
    # print alias_list, linkermap_section.gident
    for alias in alias_list:
        if alias.endswith('*'):
            alias = alias[:-1]
        if alias.endswith('.'):
            alias = alias[:-1]
        if sm.linkermap_section is not None and \
                alias == sm.linkermap_section.gident:
            continue
        # print alias, linkermap_section.gident
        if sm.linkermap_section is not None:
            sm.memory_map.aliases.register_alias(
                sm.linkermap_section.gident, alias
            )
        else:
            logging.warning("Target for alias unknown : " + alias)


def process_linkermap_line(l, sm):
    if sm.LINKERMAP_STATE == 'GOT_SECTION_NAME':
        process_linkermap_section_heading_detail_line(l, sm)
    elif sm.LINKERMAP_STATE == 'NORMAL':
        for key, regex in iteritems(re_linkermap):
            if regex.match(l):
                if key == 'LOAD':
                    process_linkermap_load_line(l, sm)
                    return
                if key == 'DEFN_ADDR':
                    process_linkermap_defn_addr_line(l, sm)
                    return
                if key == 'SECTION_HEADINGS':
                    process_linkermap_section_headings_line(l, sm)
                    return
                logging.error(
                    "Unhandled line in linkerm : {0}".format(l.strip()))
    elif sm.LINKERMAP_STATE == 'IN_SECTION':
        if sm.linkermap_symbol is not None:
            if re_linkermap['SYMBOLDETAIL'].match(l):
                process_linkermap_section_detail_line(l, sm)
                return
        if re_linkermap['SECTION_HEADINGS'].match(l):
            process_linkermap_section_headings_line(l, sm)
            return
        if re_linkermap['FILL'].match(l):
            process_linkermap_fill_line(l, sm)
            return
        if re_linkermap['LINKALIASES'].match(l):
            process_linkaliases_line(l, sm)
            return
        if re_linkermap['SYMBOL'].match(l):
            process_linkermap_symbol_line(l, sm)
            return
        if re_linkermap['SYMBOLONLY'].match(l):
            process_linkermap_symbolonly_line(l, sm)
            return
        logging.warning("Unhandled line in section : {0}".format(l.strip()))
    return None


def cleanup_and_pack_map(sm):
    for node in sm.memory_map.root.all_nodes():
        if len(node.children) > 0:
            logging.warning('Force clearing leaf size for intermediate node'
                            ' : {0}'.format(node.gident))
            node.osize = '0x0'
            node.fillsize = 0


def process_map_file(fname, profile=None):
    if profile is None:
        profile = get_profile('default')
    elif profile == 'auto':
        profile = guess_profile(fname)
    sm = GCCMemoryMapParserSM(ctx=profile)
    with open(fname) as f:
        for line in f:
            if not line.strip():
                continue
            rval = check_line_for_heading(line)
            if rval is not None:
                sm.state = rval
            else:
                if sm.state == 'IN_DEPENDENCIES':
                    process_dependencies_line(line, sm)
                elif sm.state == 'IN_COMMON_SYMBOLS':
                    process_common_symbols_line(line, sm)
                elif sm.state == 'IN_DISCARDED_INPUT_SECTIONS':
                    process_discarded_input_section_line(line, sm)
                elif sm.state == 'IN_MEMORY_CONFIGURATION':
                    process_memory_configuration_line(line, sm)
                elif sm.state == 'IN_LINKER_SCRIPT_AND_MEMMAP':
                    process_linkermap_line(line, sm)
    cleanup_and_pack_map(sm)
    return sm


if __name__ == '__main__':
    from .cli import main
    main()




# Copyright (C) 2015 Quazar Technologies Pvt. Ltd.
#               2015 Chintalagiri Shashank
#
# This file is part of fpv-gcc.
#
# fpv-gcc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fpv-gcc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fpv-gcc.  If not, see <http://www.gnu.org/licenses/>.


import logging

from fpvgcc.datastructures.ntreeSize import SizeNTree, SizeNTreeNode


class LinkAliases(object):
    def __init__(self):
        self._aliases = {}

    def register_alias(self, target, alias):
        if alias in self._aliases.keys():
            if target != self._aliases[alias]:
                logging.warning("Alias Collision : {0} :: {1} ; {2}"
                                "".format(alias, target, self._aliases[alias]))
        else:
            self._aliases[alias] = target

    def encode(self, name):
        for key in self._aliases.keys():
            if name.startswith(key):
                # if alias.startswith(linkermap_section.gident):
                # alias = alias[len(linkermap_section.gident):]
                return self._aliases[key] + name
        return name

    def __repr__(self):
        return '\n'.join(["{0:>38} -> {1:<38}".format(a, t)
                          for (a, t) in sorted(self._aliases.items(),
                                               key=lambda x: x[1])])


class GCCMemoryMapNode(SizeNTreeNode):
    _leaf_property = '_size'
    _ident_property = 'name'

    def __init__(self, parent=None, node_t=None,
                 name=None, address=None, size=None, fillsize=None,
                 arfile=None, objfile=None, arfolder=None):
        super(GCCMemoryMapNode, self).__init__(parent, node_t)
        self._size = None
        if name is None:
            if objfile is not None:
                name = objfile
            else:
                name = ""
        self.name = name
        if address is not None:
            self._address = int(address, 16)
        else:
            self._address = None
        if size is not None:
            self._defsize = int(size, 16)
        else:
            self._defsize = None
        self.arfolder = arfolder
        self.arfile = arfile
        self.objfile = objfile
        self._fillsize = None
        self.fillsize = fillsize

    @property
    def ctx(self):
        return self.parent.ctx

    @property
    def address(self):
        if self._address is not None:
            return format(self._address, '#010x')
        else:
            return ""

    @address.setter
    def address(self, value):
        self._address = int(value, 16)

    def contains_address(self, addr):
        if not self.address or self.region in ['DISCARDED', 'UNDEF']:
            return False
        if int(self.address, 0) <= int(addr, 0) < \
                (int(self.address, 0) + self.size):
            return True
        return False

    @property
    def defsize(self):
        return self._defsize

    @defsize.setter
    def defsize(self, value):
        self._defsize = int(value, 16)

    @property
    def osize(self):
        return self._size

    @osize.setter
    def osize(self, value):
        if len(self.children):
            logging.warning("Setting leaf property at a node which "
                            "has children : {0}".format(self.gident))
        newsize = int(value, 16)
        if self._size is not None:
            if newsize != self._size:
                logging.warning(
                    "Overwriting leaf property at node : {0} :: {1} -> {2}"
                    "".format(self.gident, self._size, newsize)
                )
            else:
                logging.warning("Possibly missing leaf node "
                                "with same name : {0}".format(self.gident))
        self._size = newsize

    @property
    def fillsize(self):
        return self._fillsize

    @fillsize.setter
    def fillsize(self, value):
        if value:
            try:
                self._fillsize = int(value, 0)
            except TypeError:
                self._fillsize = int(value)
        else:
            self._fillsize = 0

    def add_child(self, newchild=None, name=None,
                  address=None, size=None, fillsize=0,
                  arfile=None, objfile=None, arfolder=None):
        if newchild is None:
            nchild = GCCMemoryMapNode(name=name, address=None, size=None,
                                      fillsize=0, arfile=None, objfile=None,
                                      arfolder=None)
            newchild = super(GCCMemoryMapNode, self).add_child(nchild)
        else:
            newchild = super(GCCMemoryMapNode, self).add_child(newchild)
        return newchild

    def push_to_leaf(self):
        if not self.objfile:
            logging.warning("No objfile defined. Can't push to leaf : "
                            "{0}".format(self.gident))
            return self
        for child in self.children:
            # TODO This probably discards data which could be preserved
            if child.name == self.objfile.replace('.', '_'):
                return self
        newleaf = self.add_child(name=self.objfile.replace('.', '_'))
        if self._defsize is not None:
            newleaf.defsize = hex(self._defsize)
        if self._address is not None:
            newleaf.address = hex(self._address)
        if self.fillsize is not None:
            newleaf.fillsize = self.fillsize
        if self.objfile is not None:
            newleaf.objfile = self.objfile
        if self.arfile is not None:
            newleaf.arfile = self.arfile
        if self.arfolder is not None:
            newleaf.arfolder = self.arfolder
        newleaf.osize = hex(self._size)

        self._size = None
        self._defsize = None
        if not self.is_toplevelnode:
            self._address = None
        self.fillsize = None
        self.objfile = None
        self.arfile = None
        self.arfolder = None

        return newleaf

    @property
    def leafsize(self):
        if self.fillsize is not None:
            if self._size is not None:
                return self._size + self.fillsize
            else:
                return self.fillsize
        return self._size

    @leafsize.setter
    def leafsize(self, value):
        raise AttributeError

    @property
    def region(self):
        if not isinstance(self.parent, GCCMemoryMap) and \
                self.parent.region == 'DISCARDED':
            return 'DISCARDED'
        # Suppressed root identifiers for MSP430 GCC. A better mechanism to
        # provide user access to manipulate this set is needed.
        if self.ctx and self.name in self.ctx.suppressed_names:
            return 'DISCARDED'
        if self._address is None:
            return 'UNDEF'
        if self._address == 0:
            return "DISCARDED"
        for region in self.tree.memory_regions:
            if self._address in region:
                if region.name in self.ctx.suppressed_regions:
                    return 'DISCARDED'
                return region.name
        raise ValueError(self._address)

    @property
    def is_leaf_property_set(self):
        return self._is_leaf_property_set

    def __repr__(self):
        r = '{0:.<60}{1:<15}{2:>10}{6:>10}{3:>10}    {5:<15}{4}' \
            ''.format(self.gident, self.address or '', self.defsize or '',
                      self.size or '', self.objfile or '', self.region,
                      self._size or '')
        return r


class GCCMemoryMap(SizeNTree):
    # This really needs to be cleaned up. Most of the code here is pretty
    # much an example of what NOT to do.
    node_t = GCCMemoryMapNode
    collapse_vectors = True

    def __init__(self, ctx):
        self.ctx = ctx
        self.memory_regions = []
        self._vector_regions = []
        self._vector_sections = []
        self.aliases = LinkAliases()
        super(GCCMemoryMap, self).__init__()

    @property
    def used_regions(self):
        ur = ['UNDEF']
        if self.collapse_vectors:
            self._vector_regions = []
            ur.append('VEC')
        for node in self.root.all_nodes():
            if node.region not in ur:
                if self.collapse_vectors:
                    if 'VEC' not in node.region:
                        ur.append(node.region)
                    elif node.region not in self._vector_regions:
                        self._vector_regions.append(node.region)
                else:
                    ur.append(node.region)
        ur.remove('UNDEF')
        ur.remove('DISCARDED')
        return ur

    @property
    def used_objfiles(self):
        of = []
        for node in self.root.all_nodes():
            if node.objfile is None and node.leafsize \
                    and node.region not in ['DISCARDED', 'UNDEF']:
                logging.warning(
                    "Object unaccounted for : {0:<40} {1:<15} {2:>5}"
                    "".format(node.gident, node.region, str(node.leafsize))
                )
                continue
            if node.objfile not in of:
                of.append(node.objfile)
        return of

    def arfile_objfiles(self, arfile):
        of = []
        for node in self.root.all_nodes():
            if node.leafsize and node.region not in ['DISCARDED', 'UNDEF']:
                continue
            if node.arfile == arfile and node.objfile not in of:
                of.append(node.objfile)
        return of

    @property
    def used_arfiles(self):
        af = []
        for node in self.root.all_nodes():
            if node.arfile is None and node.leafsize \
                    and node.region not in ['DISCARDED', 'UNDEF']:
                logging.warning(
                    "Object unaccounted for : {0:<40} {1:<15} {2:>5}"
                    "".format(node.gident, node.region, str(node.leafsize))
                )
                continue
            if node.arfile not in af:
                af.append(node.arfile)
        return af

    @property
    def used_files(self):
        af = []
        of = []
        for node in self.root.all_nodes():
            if node.arfile is None and node.leafsize \
                    and node.region not in ['DISCARDED', 'UNDEF']:
                if node.objfile is None and node.leafsize \
                        and node.region not in ['DISCARDED', 'UNDEF']:
                    logging.warning(
                        "Object unaccounted for : {0:<40} {1:<15} {2:>5}"
                        "".format(node.gident, node.region, str(node.leafsize))
                    )
                    continue
                else:
                    if node.objfile not in of:
                        of.append(node.objfile)
            if node.arfile not in af:
                af.append(node.arfile)
        if None in af:
            af.remove(None)
        if None in of:
            of.remove(None)
        return of, af

    @property
    def used_sections(self):
        sections = [node.gident for node in self.top_level_nodes
                    if node.size > 0 and
                    node.region not in ['DISCARDED', 'UNDEF']]
        sections += [node.gident for node in
                     sum([n.children for n in self.top_level_nodes
                          if n.region == 'UNDEF'], [])
                     if node.region != 'DISCARDED' and node.size > 0]
        if self.collapse_vectors:
            self._vector_sections = [x for x in sections if 'vec' in x]
            sections = ['.*vec*'] + [x for x in sections if 'vec' not in x]
        return sections

    @property
    def all_symbols(self):
        asym = []
        for node in self.root.all_nodes():
            if node.name not in asym:
                asym.append(node.name)
        return asym

    def symbols_from_file(self, lfile):
        fsym = []
        for node in self.root.all_nodes():
            if node.name not in fsym and lfile in [node.objfile, node.arfile]:
                fsym.append(node.name)
        return fsym

    def get_symbol_fp(self, symbol):
        r = []
        for rgn in self.used_regions:
            r.append(self.get_symbol_fp_rgn(symbol, rgn))
        return r

    def get_symbol_fp_rgn(self, symbol, region):
        if self.collapse_vectors and region == "VEC":
            return self.get_symbol_fp_rgnvec(symbol)
        rv = 0
        for node in self.root.all_nodes():
            if node.name == symbol:
                if node.region == region:
                    if node.leafsize is not None:
                        rv += node.leafsize
        return rv

    def get_symbol_fp_rgnvec(self, symbol):
        rv = 0
        for node in self.root.all_nodes():
            if node.name == symbol:
                if 'VEC' in node.region:
                    if node.leafsize is not None:
                        rv += node.leafsize
        return rv

    def get_objfile_fp(self, objfile):
        r = []
        for rgn in self.used_regions:
            r.append(self.get_objfile_fp_rgn(objfile, rgn))
        return r

    def get_objfile_fp_rgn(self, objfile, region):
        if self.collapse_vectors and region == "VEC":
            return self.get_objfile_fp_rgnvec(objfile)
        rv = 0
        for node in self.root.all_nodes():
            if node.objfile == objfile:
                if node.region == region:
                    if node.leafsize is not None:
                        rv += node.leafsize
        return rv

    def get_objfile_fp_rgnvec(self, objfile):
        rv = 0
        for node in self.root.all_nodes():
            if node.objfile == objfile:
                if 'VEC' in node.region:
                    if node.leafsize is not None:
                        rv += node.leafsize
        return rv

    def get_objfile_fp_secs(self, objfile):
        r = []
        for section in self.used_sections:
            r.append(self.get_objfile_fp_sec(objfile, section))
        return r

    def get_arfile_fp_secs(self, arfile):
        r = []
        for section in self.used_sections:
            r.append(self.get_arfile_fp_sec(arfile, section))
        return r

    def get_objfile_fp_sec(self, objfile, section):
        if section == '.*vec*':
            return self.get_objfile_fp_secvec(objfile)
        rv = 0
        for node in self.get_node(section).all_nodes():
            if node.objfile == objfile:
                if node.leafsize is not None:
                    rv += node.leafsize
        return rv

    def get_objfile_fp_secvec(self, objfile):
        rv = 0
        for section in self._vector_sections:
            for node in self.get_node(section).all_nodes():
                if node.objfile == objfile:
                    if node.leafsize is not None:
                        rv += node.leafsize
        return rv

    def get_arfile_fp_sec(self, arfile, section):
        if section == '.*vec*':
            return self.get_objfile_fp_secvec(arfile)
        rv = 0
        for node in self.get_node(section).all_nodes():
            if node.arfile == arfile:
                if node.leafsize is not None:
                    rv += node.leafsize
        return rv

    def get_arfile_fp_secvec(self, arfile):
        rv = 0
        for section in self._vector_sections:
            for node in self.get_node(section).all_nodes():
                if node.arfile == arfile:
                    if node.leafsize is not None:
                        rv += node.leafsize
        return rv

    def get_arfile_fp(self, arfile):
        r = []
        for rgn in self.used_regions:
            r.append(self.get_arfile_fp_rgn(arfile, rgn))
        return r

    def get_arfile_fp_rgn(self, arfile, region):
        if self.collapse_vectors and region == "VEC":
            return self.get_arfile_fp_rgnvec(arfile)
        rv = 0
        for node in self.root.all_nodes():
            if node.arfile == arfile:
                if node.region == region:
                    if node.leafsize is not None:
                        rv += node.leafsize
        return rv

    def get_arfile_fp_rgnvec(self, arfile):
        rv = 0
        for node in self.root.all_nodes():
            if node.arfile == arfile:
                if 'VEC' in node.region:
                    if node.leafsize is not None:
                        rv += node.leafsize
        return rv


class MemoryRegion(object):
    def __init__(self, name, origin, size, attribs):
        self.name = name
        self.origin = int(origin, 16)
        self.size = int(size, 16)
        self.attribs = attribs

    def __repr__(self):
        r = '{0:.<20}{1:>20}{2:>20}   {3:<20}' \
            ''.format(self.name, format(self.origin, '#010x'),
                      self.size or '', self.attribs)
        return r

    def __contains__(self, value):
        if not isinstance(value, int):
            value = int(value, 16)
        if self.origin <= value < (self.origin + self.size):
            return True
        else:
            return False





from fpvgcc.fpv import GCCMemoryMapParserSM
from fpvgcc.gccMemoryMap import GCCMemoryMap
from .vectors import example_map


def test_process_map(example_map):
    mm, vectors = example_map
    assert isinstance(mm, GCCMemoryMapParserSM)
    assert isinstance(mm.memory_map, GCCMemoryMap)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2017 Chintalagiri Shashank
#
# This file is part of fpv-gcc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Docstring for cli
"""

import argparse
import logging
from prettytable import PrettyTable

from .fpv import process_map_file
from .profiles import profiles
from .profiles import get_profile
from .profiles.guess import guess_profile


def _build_table_header(cols, rowtitle):
    tbl = PrettyTable([rowtitle] + cols + ['TOTAL'])
    tbl.align[rowtitle] = 'l'
    for heading in cols:
        tbl.align[heading] = 'r'
    tbl.align['TOTAL'] = 'r'
    tbl.padding_width = 1
    totals = [0] * (len(cols) + 1)
    return tbl, totals


def _add_row(tbl, rowtitle, row, totals=None):
    total = sum(row)
    tbl.add_row([rowtitle] + [x or '' for x in row] + [total])
    if totals:
        totals = [totals[idx] + row[idx] for idx in range(len(row))]
        return totals


def _add_totals_row(tbl, totals):
    tbl.add_row(['TOTALS'] + totals + [''])


def _render_table(tbl):
    print(tbl.get_string(sortby='TOTAL', reversesort=True,
                         sort_key=lambda x: x[-1] or 0))


def print_symbol_fp(mm, lfile='all'):
    cols = mm.used_regions
    tbl, totals = _build_table_header(cols, 'SYMBOL')

    if lfile == 'all':
        symbols = mm.all_symbols
    else:
        symbols = mm.symbols_from_file(lfile)

    for symbol in symbols:
        nextrow = mm.get_symbol_fp(symbol)
        if not sum(nextrow):
            continue
        totals = _add_row(tbl, symbol, nextrow, totals)

    _add_totals_row(tbl, totals)
    _render_table(tbl)


def print_objfile_fp(mm, arfile='all'):
    cols = mm.used_regions
    tbl, totals = _build_table_header(cols, 'OBJFILE')

    if arfile == 'all':
        objfiles = mm.used_objfiles
    else:
        objfiles = mm.arfile_objfiles(arfile)

    for objfile in objfiles:
        nextrow = mm.get_objfile_fp(objfile)
        totals = _add_row(tbl, objfile, nextrow, totals)

    _add_totals_row(tbl, totals)
    _render_table(tbl)


def print_arfile_fp(mm):
    cols = mm.used_regions
    tbl, totals = _build_table_header(cols, 'ARFILE')

    for arfile in mm.used_arfiles:
        nextrow = mm.get_arfile_fp(arfile)
        totals = _add_row(tbl, arfile, nextrow, totals)

    _add_totals_row(tbl, totals)
    _render_table(tbl)


def print_file_fp(mm):
    cols = mm.used_regions
    tbl, totals = _build_table_header(cols, 'FILE')

    objfiles, arfiles = mm.used_files

    for objfile in objfiles:
        nextrow = mm.get_objfile_fp(objfile)
        totals = _add_row(tbl, objfile, nextrow, totals)

    for arfile in arfiles:
        nextrow = mm.get_arfile_fp(arfile)
        totals = _add_row(tbl, arfile, nextrow, totals)

    _add_totals_row(tbl, totals)
    _render_table(tbl)


def print_sectioned_fp(mm):
    cols = mm.used_sections
    tbl, totals = _build_table_header(cols, 'FILE')

    arfiles = []
    objfiles = mm.used_objfiles
    # objfiles, arfiles = mm.used_files

    for objfile in objfiles:
        nextrow = mm.get_objfile_fp_secs(objfile)
        totals = _add_row(tbl, objfile, nextrow, totals)

    for arfile in arfiles:
        nextrow = mm.get_arfile_fp_secs(arfile)
        totals = _add_row(tbl, arfile, nextrow, totals)

    _add_totals_row(tbl, totals)
    _render_table(tbl)


def print_files_list(fl):
    for f in sorted(set(fl)):
        if f:
            print(f)


def print_loaded_files(sm):
    for s in sorted(set(sm.loaded_files)):
        print(s)


def print_aliases(sm):
    print(sm.memory_map.aliases)


def _get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('mapfile',
                        help='GCC generated Map file to analyze.')
    parser.add_argument('-v', '--verbose',
                        help='Include detailed warnings in the output.',
                        action='count', default=0)
    parser.add_argument('-p', '--profile', metavar='PROFILE',
                        choices=profiles.keys(), default='auto')
    action = parser.add_mutually_exclusive_group(required=True)
    action.add_argument('--sar', action='store_true',
                        help='Print summary of usage per included file.')
    action.add_argument('--sobj', metavar='ARFILE',
                        help="Print summary of usage per included object "
                             "file. Specify '.a' filename or 'all'.")
    action.add_argument('--ssym', metavar='FILE',
                        help="Print summary of usage per included symbol. "
                             "Specify '.ar' or '.o' filename or 'all'.")
    action.add_argument('--ssec', action='store_true',
                        help='Print sectioned summary of usage.')
    action.add_argument('--lmap', metavar='ROOT',
                        help="Print descendent nodes in the linker map. "
                             "Specify node for which descendents should be "
                             "printed, or 'root'.")
    action.add_argument('--lobj', metavar='OBJFILE',
                        help="Print nodes from the specified obj file.")
    action.add_argument('--lar', metavar='ARFILE',
                        help="Print nodes from the specified ar file.")
    action.add_argument('--uf', action='store_true',
                        help='Print list of all used files.')
    action.add_argument('--uarf', action='store_true',
                        help='Print list of used ar files.')
    action.add_argument('--uobjf', action='store_true',
                        help='Print list of used object files.')
    action.add_argument('--uregions', action='store_true',
                        help='Print list of used regions.')
    action.add_argument('--usections', action='store_true',
                        help='Print list of used sections.')
    action.add_argument('--lfa', action='store_true',
                        help='Print list of loaded files.')
    action.add_argument('--la', action='store_true',
                        help='Print list of detected aliases.')
    action.add_argument('--addr', metavar='ADDRESS',
                        help='Describe contents at specified address.')
    return parser


def main():
    parser = _get_parser()
    args = parser.parse_args()
    if args.verbose == 0:
        logging.basicConfig(level=logging.ERROR)
    elif args.verbose == 1:
        logging.basicConfig(level=logging.WARNING)
    elif args.verbose == 2:
        logging.basicConfig(level=logging.INFO)
    elif args.verbose == 3:
        logging.basicConfig(level=logging.DEBUG)
    if args.profile == 'auto':
        pname = guess_profile(args.mapfile)
    else:
        pname = args.profile
    profile = get_profile(pname)
    state_machine = process_map_file(args.mapfile, profile=profile)
    if args.sar:
        print_file_fp(state_machine.memory_map)
    elif args.sobj:
        print_objfile_fp(state_machine.memory_map, arfile=args.sobj)
    elif args.ssym:
        print_symbol_fp(state_machine.memory_map, lfile=args.ssym)
    elif args.ssec:
        print_sectioned_fp(state_machine.memory_map)
    elif args.uf:
        ol, al = state_machine.memory_map.used_files
        print_files_list(ol + al)
    elif args.uarf:
        print_files_list(state_machine.memory_map.used_arfiles)
    elif args.uobjf:
        print_files_list(state_machine.memory_map.used_objfiles)
    elif args.usections:
        print_files_list(state_machine.memory_map.used_sections)
    elif args.uregions:
        print_files_list(state_machine.memory_map.used_regions)
    elif args.lfa:
        print_files_list(state_machine.loaded_files)
    elif args.la:
        print_aliases(state_machine)
    elif args.lmap:
        if args.lmap == 'root':
            for node in state_machine.memory_map.top_level_nodes:
                print(node)
        else:
            n = state_machine.memory_map.get_node(args.lmap)
            for node in n.all_nodes():
                print(node)
    elif args.lar:
        for node in state_machine.memory_map.root.all_nodes():
            if node.arfile == args.lar:
                print(node)
    elif args.lobj:
        for node in state_machine.memory_map.root.all_nodes():
            if node.objfile == args.lobj:
                print(node)
    elif args.addr:
        for node in state_machine.memory_map.root.all_nodes():
            if node.contains_address(args.addr):
                print(node)





import re


def guess_profile(fpath):
    fp = open(fpath, "rb")
    fp.seek(-300 - 1, 2)

    rex = r"^OUTPUT\((?P<file>[\S]+)\s(?P<sig>[\S]+)\)$"
    line = fp.readlines()[-1].strip().decode()

    matches = re.search(rex, line)
    if matches:
        return matches.group('sig')
    return 'default'





class ContextBase(object):

    def __init__(self):
        self._suppressed_names = []
        self._suppressed_regions = ['*default*']

    @property
    def suppressed_names(self):
        return self._suppressed_names

    def suppress_name(self, name):
        self._suppressed_names.append(name)

    def unsuppress_name(self, name):
        while name in self._suppressed_names:
            self._suppressed_names.remove(name)

    @property
    def suppressed_regions(self):
        return self._suppressed_regions




# Copyright (C) 2015 Quazar Technologies Pvt. Ltd.
#               2015 Chintalagiri Shashank
#
# This file is part of fpv-gcc.
#
# fpv-gcc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fpv-gcc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fpv-gcc.  If not, see <http://www.gnu.org/licenses/>.


import logging
from os.path import commonprefix


class NTreeNode(object):
    _leaf_property = None
    _ident_property = None

    def __init__(self, parent=None, node_t=None):
        self.parent = parent
        if node_t is None:
            if self.parent is not None:
                node_t = self.parent.node_t
            else:
                node_t = NTreeNode
        self.node_t = node_t
        self.children = []

    @property
    def tree(self):
        if isinstance(self.parent, NTree):
            return self.parent
        else:
            return self.parent.tree

    @property
    def is_root(self):
        if isinstance(self.parent, NTree):
            return True
        else:
            return False

    @property
    def is_toplevelnode(self):
        if self.is_root:
            return False
        if isinstance(self.parent.parent, NTree):
            return True
        return False

    @property
    def is_leaf(self):
        if len(self.children) > 0:
            return False
        else:
            return True

    def add_child(self, newchild=None):
        if newchild is None:
            newchild = self.node_t(parent=self, node_t=self.node_t)
        try:
            for child in self.children:
                if child.ident == newchild.ident:
                    raise ValueError("Child with that identifier already "
                                     "exists: {0}".format(newchild.ident))
        except NotImplementedError:
            pass
        if len(self.children) == 0:
            try:
                if self._is_leaf_property_set is True:
                    logging.warning(
                        "Adding children to node which has leaf specific "
                        "information set : {0}".format(self.gident)
                    )
            except NotImplementedError:
                pass
        newchild.parent = self
        self.children.append(newchild)
        return newchild

    @property
    def _is_leaf_property_set(self):
        if self._leaf_property is None:
            raise NotImplementedError
        if getattr(self, self._leaf_property) is not None:
            return True
        return False

    @property
    def _is_ident_property_set(self):
        if not self._ident_property:
            raise NotImplementedError
        if not getattr(self, self._ident_property, None):
            return True
        return False

    @property
    def idx(self):
        if self.parent is None:
            return 'Root'
        else:
            return self.parent.children.index(self)

    @property
    def ident(self):
        if self._ident_property:
            return getattr(self, self._ident_property, '')
        else:
            return str(self.idx)

    @ident.setter
    def ident(self, value):
        if self._ident_property:
            setattr(self, self._ident_property, value)

    @property
    def has_ident(self):
        return self._ident_property is not None

    @property
    def gident(self):
        rval = self.ident
        walker = self
        while not isinstance(walker.parent, NTree):
            walker = walker.parent
            rval = walker.ident + '.' + rval
        return rval

    def get_child_by_ident(self, ident):
        for child in self.children:
            if child.ident == ident:
                return child
        raise ValueError

    def get_child_disambig(self, ident, prospective=False):
        disambig = None
        for child in self.children:
            if ':' in child.ident:
                name, d = child.ident.split(':')
                if name == ident:
                    disambig = disambig or 0
                    disambig = max(disambig, int(d))
            elif prospective and child.ident == ident:
                # print("Recommending disambig for {0};{1}
                # ".format(self.gident, ident))
                disambig = disambig or 0
        return disambig

    def get_descendent_by_ident(self, ident):
        res = self.get_child_by_ident(ident)
        if res is not None:
            return res
        for child in self.children:
            if child.is_leaf is False:
                res = child.get_descendent_by_ident(ident)
                if res is not None:
                    return res
        return ValueError

    def all_nodes(self):
        yield self
        for child in self.children:
            for node in child.all_nodes():
                yield node

    @property
    def get_top_level_ancestor(self):
        walker = self
        if walker.is_root:
            return None
        while not walker.is_toplevelnode:
            walker = walker.parent
        return walker


class NTree(object):
    node_t = NTreeNode

    def __init__(self):
        self.root = self.node_t(parent=self, node_t=self.node_t)

    @property
    def top_level_nodes(self):
        return self.root.children

    @property
    def top_level_idents(self):
        if self.root.has_ident:
            return [x.ident for x in self.top_level_nodes]
        else:
            raise AttributeError

    @property
    def top_level_gidents(self):
        return [x.gident for x in self.top_level_nodes]

    def get_node_disambig(self, gident, prospective=False):
        crumbs = gident.split('.')
        if crumbs[0] != self.root.ident:
            raise ValueError
        walker = self.root
        for crumb in crumbs[1:-1]:
            if walker.get_child_disambig(crumb) is None:
                walker = walker.get_child_by_ident(crumb)
            else:
                raise ValueError('Deep disambig found at intermediate node : '
                                 '{0}'.format(crumbs))
        return walker.get_child_disambig(crumbs[-1], prospective)

    def get_node(self, gident, create=False):
        crumbs = gident.split('.')
        if crumbs[0] != self.root.ident:
            raise ValueError
        walker = self.root
        for crumb in crumbs[1:]:
            try:
                walker = walker.get_child_by_ident(crumb)
            except ValueError:
                if create:
                    # print "Creating Node : " + crumb
                    newnode = walker.add_child()
                    if newnode.has_ident:
                        newnode.ident = crumb
                    else:
                        raise ValueError
                    walker = walker.get_child_by_ident(crumb)
                else:
                    raise ValueError
        return walker

    def get_least_common_ancestor(self, nodes):
        t = commonprefix([node.gident for node in nodes])
        k = t.rfind('.')
        return self.get_node(t[:k])





from .context import ContextBase
from .gcc_msp430 import ProfileGccMsp430


def _load_profiles():
    return {
        'default': ContextBase,
        ProfileGccMsp430.id: ProfileGccMsp430,
    }


profiles = _load_profiles()


def get_profile(idn):
    if idn in profiles.keys():
        return profiles[idn]()
    else:
        return profiles['default']()




from .context import ContextBase


class ProfileGccMsp430(ContextBase):
    id = 'elf32-msp430'

    def __init__(self):
        super(ProfileGccMsp430, self).__init__()
        self._suppressed_names.extend([
                'debug_line',
                'debug_info',
                'debug_ranges',
                'debug_aranges',
                'debug_loc',
                'debug_str',
                'debug_frame',
                'debug_abbrev',
                'comment',
                'MSP430'
        ])
